## DTHybrid  disponibile all'indirizzo http://alpha.dmi.unict.it/dtweb/dthybrid.php
library("DTHybrid")
## Tutto ci di cui si ha bisogno  disponibile nel pacchetto recommenderlab
## da installare tramite install.packages("recommenderlab", dependencies=TRUE)
library("recommenderlab")

#####################################################################################################
# Definizione di alcune funzioni d'utilit
readkey <- function() {
  cat ("Press [enter] to continue")
  line <- readline()
}

#####################################################################################################
# Esempio di costruzione degli oggetti del modello

## Costruisco una matrice che rappresenta un sistema con 5 utenti e 10 oggetti
## i valori sono casuali scelti tra 0 e 5 (se l'utente ha votato)
## NA per i rating sconosciuti
ratings <- matrix(sample(c(as.numeric(0:5), NA), 50, replace=TRUE, prob=c(rep(.4/6,6),.6)), 
                  ncol=10, dimnames=list(paste0("u", 1:5),
                                         paste0("i", 1:10)))
print(ratings)

readkey()

## Costruisco l'oggetto che rappresenta la matrice di Rating come richiesto da reccomenderlab
## basta fare un cast all'oggetto realRatingMatrix
utility.matrix <- as(ratings, "realRatingMatrix") #se i valori erano binari "binaryRatingMatrix"
print(utility.matrix)

readkey()

## Per visualizzare l'oggetto basta fare un cast a lista o data.frame
print(as(utility.matrix, "list"))

readkey()

#####################################################################################################
# Esempio di normalizzazione

## Per normalizzare i rating di una realRatingMatrix basta usare la funzione normalize
utility.matrix.normalized <- normalize(utility.matrix)

## Con plot(image(...)) visualizzo graficamente una realRatingMatrix o binaryRatingMatrix
plot(image(utility.matrix, main = "Raw Ratings"))
readkey()

## La differenza tra quella normalizzata e non normalizzata emerge facilmente
## visualizzando entrambe le matrici
plot(image(utility.matrix.normalized, main = "Normalized Ratings"))
readkey()

#####################################################################################################
# Carico dei dati di esempio

data(Jester5k)
print(Jester5k)

#####################################################################################################
# User-Based Collaborative Filtering

## Per calcolare recommendation occorre addestrare un modello di predizione
## tramite il comando Recommender indicando l'oggetto contenente la matrice dei dati
## e l'algoritmo da usare.
## In questo caso si stanno usando 1000 utenti del dataset Jester5k
## UBCF = user-based collaborative filtering
r <- Recommender(Jester5k[1:1000], method = "UBCF")
print(r)

readkey()

## Addrestrato il modello si pu eseguire una predizione usando il comando predict
## n indica il numero massimo di oggetti da restituire nella predizione
recom <- predict(r, Jester5k[1001:1002], n=5)
print(recom)
## Per mostrare i risultati della predizione basta convertire l'oggetto restituito
## da predict in una lista
print(as(recom, "list"))

readkey()

## Se si  interessati anche ai rating calcolati dal predittore occorre specificare
## il parametro type
recom <- predict(r, Jester5k[1001:1002], type="ratings")
## Per visualizzare i rating ? quindi necessario convertire l'oggetto in una matrice
print(as(recom, "matrix")[,1:10])

readkey()

#####################################################################################################
# Item-Based Collaborative Filtering

## IBCF = item-based collaborative filtering
r <- Recommender(Jester5k[1:1000], method = "IBCF")
print(r)

readkey()

recom <- predict(r, Jester5k[1001:1002], n=5)
print(recom)
print(as(recom, "list"))

readkey()

recom <- predict(r, Jester5k[1001:1002], type="ratings")
print(as(recom, "matrix")[,1:10])

readkey()

#####################################################################################################
# SVD-based collaborative Filtering

## SVD = SVD approximated collaborative filtering
r <- Recommender(Jester5k[1:1000], method = "SVD")
print(r)

readkey()

recom <- predict(r, Jester5k[1001:1002], n=5)
print(recom)
print(as(recom, "list"))

readkey()

recom <- predict(r, Jester5k[1001:1002], type="ratings")
print(as(recom, "matrix")[,1:10])

readkey()

#####################################################################################################
# Valutazione dei risultati

## Per valutare i risultati  innanzi tutto necessario definire uno schema di valutazione,
## ovvero un oggetto che descrive il test da eseguire, con il comando evaluationScheme.
## Il metodo scelto per questa comparazione "split" divide il dataset in due parti:
## training set contenente il 90% degli elementi (train = 0.9), e il test set con i restanti.
## "k" indica il numero di ripetizioni, "goodRating" indica la soglia minima per considerare
## un rating buono
scheme <- evaluationScheme(Jester5k[1:1000], method="split", train = 0.9, k=1, given=20, goodRating=5)

print(scheme)

readkey()

## Costruisco una lista con gli algoritmi che voglio testare indicando per ognuno il
## nome e eventuali parametri
algorithms <- list(
  "Random"=list(name="RANDOM", param=NULL),
  "User-Based CF"=list(name="UBCF", param=NULL),
  "Item-Based CF"=list(name="IBCF", param=NULL),
  "SVD-Based CF"=list(name="SVD", param=NULL)
)

## Tramite la funzione evaluate avvio il test sugli algoritmi prescelti.
## Con il parametro n indico che voglio comparare le performance scegliendo
## un numero fissato di predizioni
results <- evaluate(scheme, algorithms, n=c(1, 3, 5, 10, 15, 20))

print(results)

readkey()

## Con plot visualizzo dei grafici che indicano l'andamento delle prestazioni.
## In questo caso visualizzo una curva ROC.
plot(results, annotate=3)

readkey()

## In quest'altro caso visualizzo una curva precision/recall
plot(results, "prec/rec", annotate=3)

readkey()

#####################################################################################################
# Utilizzo di DT-Hybrid

## L'utilizzo di DT-Hybrid  estremamente semplice.
## Basta avere le tre matrici necessarie all'esecuzione del modello:
## Adiacenza, Similarit degli oggetti e Similarit? degli utenti.
## In questo caso leggo il dataset di esempio fornito con il pacchetto
data(enzyme)

View(enzyme_r) ## Questa  la matrice di adiacenza Oggetti X Utenti
print(enzyme_r[enzyme_r[,1] != 0,1])
cat("\n\n")
readkey()
View(enzyme_ds) ## Questa  la similarit tra gli utenti
readkey()
View(enzyme_ts) ## Questa  la similarit tra gli oggetti
readkey()

## DT-Hybrid  un algoritmo parallelizzato quindi istanzio un cluster
library(parallel)
cl <- makeCluster(detectCores())
## Richiamare l'algoritmo  semplicissimo basta eseguire il comando
## computeRecommendation specificando le matrici, i parametri alpha e lambda, e l'oggetto cluster
result <- computeRecommendation(enzyme_r, S=enzyme_ts, S1=enzyme_ds, alpha=0.5, lambda=0.5, cl=cl)
## Chiudo il cluster
stopCluster(cl)


## Il risultato di DT-Hybrid  una matrice che posso interrogare come voglio
## In questo esempio estraggo tutte le predizioni per il primo utente
## Per prima cosa vedo quali sono gli oggetti che il primo utente non aveva
not.rated <- which(enzyme_r[,1] == 0)
## Ne estraggo tutti i rating
predicted.rating <- result[not.rated, 1]
## Mostro solo quelli con valore diverso da zero ordinandoli dal pi grande
## al pi piccolo
print(sort(predicted.rating[predicted.rating != 0], decreasing=TRUE))
cat("\n\n")
readkey()